#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


def get_R_grid_fft ( data_controller, nr1, nr2, nr3):
  import numpy as np
  from scipy import fftpack as FFT

  arrays = data_controller.data_arrays
  attributes = data_controller.data_attributes

  nrtot = nr1*nr2*nr3

  a_vectors = arrays['a_vectors']

  arrays['R'] = np.zeros((nrtot,3), dtype=float)
  arrays['idx'] = np.zeros((nr1,nr2,nr3), dtype=int)
  arrays['Rfft'] = np.zeros((nr1,nr2,nr3,3), dtype=float)
  arrays['R_wght'] = np.ones((nrtot), dtype=float)

  for i in range(nr1):
    for j in range(nr2):
      for k in range(nr3):
        n = k + j*nr3 + i*nr2*nr3
        Rx = float(i)/float(nr1)
        Ry = float(j)/float(nr2)
        Rz = float(k)/float(nr3)
        if Rx >= 0.5: Rx=Rx-1.0
        if Ry >= 0.5: Ry=Ry-1.0
        if Rz >= 0.5: Rz=Rz-1.0
        Rx -= int(Rx)
        Ry -= int(Ry)
        Rz -= int(Rz)

        arrays['R'][n,:] = Rx*nr1*a_vectors[0,:] + Ry*nr2*a_vectors[1,:] + Rz*nr3*a_vectors[2,:]
        arrays['Rfft'][i,j,k,:] = arrays['R'][n,:]
        arrays['idx'][i,j,k] = n



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
import scipy.linalg as la
from numpy import linalg as npl
from .constants import ANGSTROM_AU

def do_berry_phase (self):

  import os

  arry,attr = self.data_controller.data_dicts()

  occupied = attr['berry_occupied']
  sub = arry['berry_sub']
  contin = attr['berry_contin']
  fname = attr['berry_fname']

  kxlim = arry['berry_kxlim']
  kylim = arry['berry_kylim']

  if occupied:
    ovr_ndim = int(attr['nelec'])
  elif sub != None:
    ovr_ndim = len(sub)
  else: ovr_ndim = arry['HRs'].shape[0]

  if attr['berry_kspace_method'] == 'path':
    # Calculate the bands
    do_berry_bands(self.data_controller)
    # Calculate the Berry/Zak phase
    phase = do_phase(self.data_controller)

    if not attr['berry_eigvals']:
      attr['berry_phase'] = phase
    else: arry['berry_phase'] = phase

    if not attr['berry_eigvals']:
      # Write 'berry_phase.dat'
      with open(os.path.join(attr['opath'],fname+'.dat'), 'w') as f:
        f.write(f'1D Berry phase (mod pi): phi = {phase/np.pi: 2.6f} \n')
    else: self.data_controller.write_bands(fname, phase)

  elif attr['berry_kspace_method'] == 'track':

    nk1,nk2 = attr['berry_nk1'],attr['berry_nk2']
    kpath_funct = attr['berry_kpath_funct']
    kpts = np.linspace(0.0,1.0,nk2)

    if not attr['berry_eigvals']:
      phase = np.zeros(kpts.shape[0])
    else:
      phase = np.zeros((kpts.shape[0],ovr_ndim))

    for ik,k in enumerate(kpts):
      attr['berry_path'],arry['berry_high_sym_points'] = kpath_funct(k)

      do_berry_bands(self.data_controller)
      phase[ik] = do_phase(self.data_controller)

    if not attr['berry_eigvals']:

      if contin: phase = berry_phase_cont(phase,phase[0])
      phase -= phase[0]

      with open(os.path.join(attr['opath'],fname+'.dat'), 'w') as f:
        f.write('# k\tphi\n')
        for ik,k in enumerate(kpts):
          f.write(f'{k: 2.6f}\t{phase[ik]: 2.6f}\n')
    else:

      if contin: phase = berry_eigvals_cont(phase,phase[0,:])

      self.data_controller.write_berry_bands(fname,phase)

    arry['berry_phase'] = phase

  elif attr['berry_kspace_method'] == 'circle':

    nk = attr['berry_nk']
    kradius = attr['berry_kradius']
    kcenter = arry['berry_kcenter']
    kz = kcenter[2]

    ang = np.linspace(0,2*np.pi,nk)

    path = []

    for ik,phi in enumerate(ang):
      kx,ky = np.cos(phi),np.sin(phi)
      kx *= kradius
      ky *= kradius

      kx += kcenter[0]
      ky += kcenter[1]

      path.append([kx,ky,kz])

    arry['berry_kq'] = np.array(path).T
    arry['berry_contour'] = np.copy(arry['berry_kq'])

    do_berry_bands(self.data_controller)
    phase = do_phase(self.data_controller)

    attr['berry_phase'] = phase

    # Write 'berry_phase.dat'
    with open(os.path.join(attr['opath'],fname+'.dat'), 'w') as f:
      f.write(f'k-space circle of radius {kradius} /Ang^-1\ncentered at k-point {kcenter}\n')
      f.write(f'Berry phase: {phase: 2.6f} \n')

  elif attr['berry_kspace_method'] == 'square':

    nk1 = attr['berry_nk1']
    nk2 = attr['berry_nk2']

    if nk1 % 2 != 0: nk1 += 1
    if nk2 % 2 != 0: nk2 += 1

    kx_points = np.linspace(kxlim[0],kxlim[1],nk1)
    ky_points = np.linspace(kylim[0],kylim[1],nk2)

    kpts = np.zeros((nk1,nk2,3))
    kgrid_centers = np.zeros((nk1-1,nk2-1,3))
    phases = np.zeros((nk1-1,nk2-1))

    for jk in range(nk2):
      for ik in range(nk1):
        kpts[ik,jk] = [kx_points[ik],ky_points[jk],0]

    for jk in range(nk2-1):
      for ik in range(nk1-1):

        k00 = kpts[ik,jk]
        k10 = kpts[ik+1,jk]
        k11 = kpts[ik+1,jk+1]
        k01 = kpts[ik,jk+1]

        arry['berry_kq'] = np.array([k00,k10,k11,k01,k00]).T
        arry['berry_contour'] = np.copy(arry['berry_kq'])
        kgrid_centers[ik,jk] = np.array([k00,k10,k11,k01,k00]).mean(axis=0)

        do_berry_bands(self.data_controller)
        phases[ik,jk] = do_phase(self.data_controller)

    arry['berry_kgrid'] = kpts
    arry['berry_kgrid_centers'] = kgrid_centers

    for i in range(phases.shape[1]):
      if i == 0: clos = phases[0,0]
      else: clos = phases[0,i-1]
      phases[:,i] = berry_phase_cont(phases[:,i],clos)

    arry['berry_phase'] = phases
    attr['berry_flux'] = arry['berry_phase'].sum()

    with open(os.path.join(attr['opath'],fname+'.dat'), 'w') as f:
      f.write(f'# xlim: ({kpts[0,0,0]},{kpts[-1,0,0]}); ylim: ({kpts[0,0,1]},{kpts[0,-1,1]})\n')
      f.write(f'# phases with shape: ({nk1-1},{nk2-1})\n')
      f.write(f'# kx,ky are mesh centers\n')
      f.write( '# kx\tky\tphi\n')
      for jk in range(nk2-1):
        for ik in range(nk1-1):
          f.write(f'{kgrid_centers[ik,jk][0]: 2.12f}\t{kgrid_centers[ik,jk][1]: 2.12f}\t{phases[ik,jk]: 2.12f}\n')

    with open(os.path.join(attr['opath'],fname+'_kgrid_corners.dat'), 'w') as f:
      f.write( '# kx\tky\n')
      for jk in range(nk2):
        for ik in range(nk1):
          f.write(f'{kpts[ik,jk][0]: 2.12f}\t{kpts[ik,jk][1]: 2.12f}\n')


def do_phase (data_controller):

  arry,attr = data_controller.data_dicts()
    
  v_kp = arry['berry_v_k']
  E_kp = arry['berry_E_k']

  nkpts = v_kp.shape[0]

  berry_eigvals = attr['berry_eigvals']

  closed = attr['berry_closed']
  method = attr['berry_method']
  sub = arry['berry_sub']
  occupied = attr['berry_occupied']

  alat = attr['alat'] / ANGSTROM_AU
  a_vectors = arry['a_vectors'] * alat
  b_vectors = arry['b_vectors'] * (1/alat)

  tau = arry['tau']
  naw = arry['naw']

  contour = arry['berry_contour']
  
  if np.allclose(contour[:,0], contour[:,-1]):
    closed = False
  
  method = method.lower()
  if method == 'berry':
    pass
  elif method == 'zak':
    closed = True   

  # assumes that occupancy does not change throughout the choosen path

  if occupied:
    sub = None
    occ_idx = np.arange(0,attr['nelec'],1,dtype=int)
    num_occupied = len(occ_idx)
    dim = num_occupied
  elif not sub is None:
    occ_idx = np.copy(sub)
    num_occupied = len(sub)
    dim = len(sub)
  else:
    occ_idx = np.arange(0,arry['HRs'].shape[0],1,dtype=int)
    num_occupied = len(occ_idx)
    dim = num_occupied
    
  prd = np.eye(dim,dtype=complex)
  ovr = np.zeros([dim,dim],dtype=complex)

  for ik in range(nkpts-1):

    jk = ik+1
    
    left_eig,left_states = E_kp[ik,:,0], v_kp[ik,:,:,0]
    right_eig,right_states = E_kp[jk,:,0], v_kp[jk,:,:,0]

    if occupied or not sub is None:

      left_states_idx = occ_idx
      right_states_idx = occ_idx

      left_occ = np.zeros((left_states.shape[0],len(left_states_idx)),dtype=complex)
      right_occ = np.zeros((right_states.shape[0],len(right_states_idx)),dtype=complex)

      for idx,occ in enumerate(left_states_idx):
        left_occ[:,idx] = left_states[:,occ]
      for idx,occ in enumerate(right_states_idx):
        right_occ[:,idx] = right_states[:,occ]
      left_states = left_occ
      right_states = right_occ

    else: pass

    ovr = np.dot(left_states.conj().T,right_states)
    Z_ovr, sig_ovr, W_dag_ovr = la.svd(ovr)
    ovr = np.matmul(Z_ovr,W_dag_ovr)

    prd = np.dot(prd,ovr)
    
  if closed:
    ik = nkpts-1
    jk = 0
    
    left_eig,left_states = E_kp[ik,:,0], v_kp[ik,:,:,0]
    right_eig,right_states = E_kp[jk,:,0], v_kp[jk,:,:,0]
    
    if method == "zak":
      axis = contour[:,1] - contour[:,0]
      axis /= np.dot(axis.T,axis) ** 0.5

      G = np.dot(axis,b_vectors) * 2 * np.pi

      orb_sites = np.repeat(tau,naw,axis=0)

      phase = np.dot(orb_sites, G).reshape(-1,1)
      
      left_states = right_states * np.exp(-1j * phase)

    if occupied or not sub is None:

      left_states_idx = occ_idx
      right_states_idx = occ_idx

      left_occ = np.zeros((left_states.shape[0],len(left_states_idx)),dtype=complex)
      right_occ = np.zeros((right_states.shape[0],len(right_states_idx)),dtype=complex)

      for idx,occ in enumerate(left_states_idx):
        left_occ[:,idx] = left_states[:,occ]
      for idx,occ in enumerate(right_states_idx):
        right_occ[:,idx] = right_states[:,occ]
      left_states = left_occ
      right_states = right_occ

    else: pass
    
    ovr = np.dot(left_states.conj().T,right_states)

    Z_ovr, sig_ovr, W_dag_ovr = la.svd(ovr)
    ovr = np.matmul(Z_ovr,W_dag_ovr)

    prd = np.dot(prd,ovr)

  if not berry_eigvals:
    det = la.det(prd)
    ret = -1.0 * np.angle(det)

    return ret

  else:
    evals,evecs = la.eig(prd)
    ret = -1.0 * np.angle(evals)
    ret = np.sort(ret)

    return ret

def bands_calc ( data_controller ):
  from .communication import scatter_full, gather_full

  arry,attr = data_controller.data_dicts()

  npool = attr['npool']
  nawf,_,nk1,nk2,nk3,nspin = arry['HRs'].shape

  kq_aux = scatter_full(arry['berry_kq'].T, npool).T
 
  Hks_aux = band_loop_H(data_controller, kq_aux)

  E_kp_aux = np.zeros((kq_aux.shape[1],nawf,nspin), dtype=float, order="C")
  v_kp_aux = np.zeros((kq_aux.shape[1],nawf,nawf,nspin), dtype=complex, order="C")

  for ispin in range(nspin):
    for ik in range(kq_aux.shape[1]):
      E_kp_aux[ik,:,ispin],v_kp_aux[ik,:,:,ispin] = la.eigh(Hks_aux[:,:,ik,ispin], b=(None), lower=False, overwrite_a=True, overwrite_b=True, turbo=True, check_finite=True)

  arry['berry_Hks'] = Hks_aux

  Hks_aux = Sks_aux = None
  return E_kp_aux, v_kp_aux


def band_loop_H ( data_controller, kq_aux ):

  arry,attr = data_controller.data_dicts()

  nksize = kq_aux.shape[1]
  nawf,_,nk1,nk2,nk3,nspin = arry['HRs'].shape

  HRs = np.reshape(arry['HRs'], (nawf,nawf,nk1*nk2*nk3,nspin), order='C')
  kdot = np.tensordot(arry['R'], 2.0j*np.pi*kq_aux, axes=([1],[0]))
  np.exp(kdot, kdot)
  Haux = np.zeros((nawf,nawf,nksize,nspin), dtype=complex, order="C")

  for ispin in range(nspin):
    Haux[:,:,:,ispin] = np.tensordot(HRs[:,:,:,ispin], kdot, axes=([2],[0]))

  kdot  = None
  return Haux

def do_berry_bands ( data_controller ):
  from mpi4py import MPI
  from .constants import ANGSTROM_AU
  from .communication import gather_full
  from .get_R_grid_fft import get_R_grid_fft

  rank = MPI.COMM_WORLD.Get_rank()

  arry,attr = data_controller.data_dicts()

  # Bohr to Angstrom
  attr['alat'] /= ANGSTROM_AU

  #--------------------------------------------
  # Compute bands on a selected path in the BZ
  #--------------------------------------------

  alat = attr['alat']
  nawf,_,nk1,nk2,nk3,nspin = arry['HRs'].shape
  nktot = nk1*nk2*nk3

  # Define real space lattice vectors
  get_R_grid_fft(data_controller, nk1, nk2, nk3)

  # Define k-point mesh for bands interpolation
  if attr['berry_kspace_method'] == 'path' or attr['berry_kspace_method'] == 'track':
    berry_kpnts_interpolation_mesh(data_controller)

  nkpi = arry['berry_kq'].shape[1]
  for n in range(nkpi):
    arry['berry_kq'][:,n] = np.dot(arry['berry_kq'][:,n], arry['b_vectors'])

  # Compute the bands along the path in the IBZ
  arry['berry_E_k'],arry['berry_v_k'] = bands_calc(data_controller)

  # Angstrom to Bohr
  attr['alat'] *= ANGSTROM_AU

def berry_kpnts_interpolation_mesh ( data_controller ):
  '''
  Get path between HSP
  Arguments:
      nk (int): total number of points in path

  Returns:
      kpoints : array of arrays kx,ky,kz
      numK    : Total no. of k-points
  '''

  from .kpnts_interpolation_mesh import get_path

  arry,attr = data_controller.data_dicts()

  dk = 0.00001
  nk,alat,ibrav = attr['berry_nk'],attr['alat'],attr['ibrav']
  a_vectors,b_vectors = arry['a_vectors'],arry['b_vectors']
  band_path,high_sym_points = attr['berry_path'],arry['berry_high_sym_points']

  bp,hsp = (band_path,high_sym_points) if len(high_sym_points)!=0 else (None,None)

  points,_ = get_path(ibrav,alat,a_vectors,dk,b_vectors,bp,hsp)

  scaled_dk = dk*(points.shape[1]/nk)

  points,path_file = get_path(ibrav,alat,a_vectors,scaled_dk,b_vectors,bp,hsp)

  data_controller.write_kpnts_path('berry_phase_kpath_points.txt', path_file, points, b_vectors)

  arry['berry_kq'] = points
  arry['berry_contour'] = np.copy(arry['berry_kq'])

def no_2pi(x,ref):
  "Make x as close to clos by adding or removing 2pi"

  while abs(ref-x) > np.pi:
    if ref-x > np.pi:
      x += 2.0*np.pi
    elif ref-x < -1.0*np.pi:
      x -= 2.0*np.pi

  return x

def berry_phase_cont ( pha, clos ):
  '''
  Reads in 1d array of numbers *pha* and makes sure that they are
  continuous, i.e., that there are no jumps of 2pi. First number is
  made as close to *clos* as possible.
  '''
  ret = np.copy(pha)

  # go through entire list and "iron out" 2pi jumps
  for i in range(len(ret)):
    # which number to compare to
    if i == 0: cmpr = clos
    else: cmpr = ret[i-1]
    # make sure there are no 2pi jumps
    ret[i] = no_2pi(ret[i],cmpr)

  return ret

def berry_eigvals_cont(arr_pha,clos):
    """Reads in 2d array of phases *arr_pha* and makes sure that they
    are continuous along first index, i.e., that there are no jumps of
    2pi. First array of phasese is made as close to *clos* as
    possible."""

    ret = np.zeros_like(arr_pha)

    # go over all points
    for i in range(arr_pha.shape[0]):
        # which phases to compare to
        if i == 0: cmpr = clos
        else: cmpr = ret[i-1,:]
        # remember which indices are still available to be matched
        avail=list(range(arr_pha.shape[1]))
        # go over all phases in cmpr[:]
        for j in range(cmpr.shape[0]):
            # minimal distance between pairs
            min_dist = 1.0E10
            # closest index
            best_k = None
            # go over each phase in arr_pha[i,:]
            for k in avail:
                cur_dist = np.abs(np.exp(1.0j*cmpr[j])-np.exp(1.0j*arr_pha[i,k]))
                if cur_dist <= min_dist:
                    min_dist = cur_dist
                    best_k = k
            # remove this index from being possible pair later
            avail.pop(avail.index(best_k))
            # store phase in correct place
            ret[i,j] = arr_pha[i,best_k]
            # make sure there are no 2pi jumps
            ret[i,j] = no_2pi(ret[i,j],cmpr[j])

    return ret







report_pre_reqs = 'SUGGESTION: %s must be called before %s'

module_pre_reqs = { 'add_external_fields'      : ['projectability'],\
                    'pao_hamiltonian'          : ['projectability'],\
                    'z2_pack'                  : ['pao_hamiltonian'],\
                    'atomic_orbitals'           : ['pao_hamiltonian'], \
                    'doubling_Hamiltonian'     : ['pao_hamiltonian'],\
                    'cutting_Hamiltonian'      : ['pao_hamiltonian'],\
                    'wave_function_projection' : ['pao_hamiltonian'],\
                    'bands'                    : ['pao_hamiltonian'],\
                    'topology'                 : ['bands', 'spin_operator'],\
                    'interpolated_hamiltonian' : ['pao_hamiltonian'],\
                    'pao_eigh'                 : ['pao_hamiltonian'],\
                    'dos'                      : ['pao_eigh'],\
                    'fermi_surface'            : ['pao_eigh'],\
                    'spin_texture'             : ['pao_eigh'],
                    'gradient_and_momenta'     : ['pao_eigh'],\
                    'ipr'                      : ['pao_eigh'],\
                    'berry_phase'              : ['pao_hamiltonian'],\
                    'real_space_wfc'           : ['gradient_and_momenta', 'atomic_orbitals'], \
                    'adaptive_smearing'        : ['gradient_and_momenta'],\
                    'anomalous_Hall'           : ['gradient_and_momenta'],\
                    'spin_Hall'                : ['gradient_and_momenta', 'spin_operator'],\
                    'transport'                : ['gradient_and_momenta'],\
                    'dielectric_tensor'        : ['gradient_and_momenta'] }


key_error_strings = { 'U' : 'Projections must be computed before projectability. Call \'read_atomic_proj_QE\' to read projections from the output of projwfc.x',\
                      'HRs'  : 'HRs must first be calculated with \'build_pao_hamiltonian\'',\
                      'd2Ed2k'  : 'd2Ed2k must first be calculated using band_curvature set to True in \'gradient_and_momenta\'',\
                      'Hks'  : 'Hks must first be calculated with \'build_pao_hamiltonian\''}



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

def add_ext_field ( data_controller ):
  import numpy as np
  from .constants import ANGSTROM_AU

  arrays = data_controller.data_arrays
  attributes = data_controller.data_attributes

  nawf,_,nk1,nk2,nk3,nspin = arrays['HRs'].shape
  arrays['HRs'] = np.reshape(arrays['HRs'], (nawf,nawf,nk1*nk2*nk3,nspin), order='C')

  l=0
  natoms = attributes['natoms']
  nwf = nawf//natoms
  tau_wf = np.zeros((nawf,3), dtype=float)
  for n in range(attributes['natoms']):
    for i in range(nwf):
      tau_wf[l,:] = arrays['tau'][n,:]
      l += 1

  tau_wf /= ANGSTROM_AU

  if arrays['Efield'].any() != 0.0:
    for n in range(nawf):
      arrays['HRs'][n,n,0,:] -= arrays['Efield'].dot(tau_wf[n,:])

  if arrays['Bfield'].any() != 0.0:
    field = arrays['Bfield'][0] * arrays['Sj'][0] + arrays['Bfield'][1] * arrays['Sj'][1] + arrays['Bfield'][2] * arrays['Sj'][2]
    for n in range(nawf):
      for m in range(nawf):
        arrays['HRs'][n,m,0,:] -= field[n,m]

  if arrays['HubbardU'].any() != 0:
    for n in range(nawf):
      arrays['HRs'][n,n,0,:] -= arrays['HubbardU'][n]/2.0

  arrays['HRs'] = np.reshape(arrays['HRs'], (nawf,nawf,nk1,nk2,nk3,nspin), order='C')



# 
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

# Physical constants

K_BOLTZMAN_SI    = 1.38066e-23        # J K^-1
K_BOLTZMAN_AU    = 3.1667e-6          # Hartree K^-1
K_BOLTZMAN_M1_AU = 315795.26e0        # Hartree^-1 K
FACTEM           = 315795.26e0        # Hartree^-1 K
H_OVER_TPI       = 1.054571e-34       # J sec

BOHR_RADIUS_SI   = 0.529177e-10       # m
BOHR_RADIUS_CM   = 0.529177e-8        # cm
BOHR_RADIUS_ANGS = 0.529177e0         # angstrom
ELECTRONMASS_SI  = 9.10953e-31        # Kg
ELECTRONMASS_UMA = 5.4858e-4          # uma

ELECTRONVOLT_SI  = 1.6021892e-19      # J
UMA_SI           = 1.66057e-27        # Kg
DEBYE_SI         = 3.33564e-30        # Coulomb meter
DEBYE_AU         = 0.393427228        # e * Bohr
AU_TO_VMM1       = 36.3609*10e10      # V/m
ANGSTROM_AU      = 1.889727e0         # au
AU_TO_OHMCMM1    = 46000.0e0          # (ohm cm)^-1
AU_KB            = 294210.0e0         # Kbar
KB_AU            = 1.0e0/294210.0e0   # au
AU               = 27.211652e0        # eV
RYD              = 13.605826e0        # eV
SCMASS           = 1822.89e0          # uma to au
UMA_AU           = 1822.89e0          # au
AU_TERAHERTZ     = 2.418e-5           # THz
TERAHERTZ        = 2.418e-5           # from au to THz
AU_SEC           = 2.4189e-17         # sec

EPS0             = 1.0/(4.0 * 3.14159265358979323846 ) # vacuum dielectric constant in Ry
RYTOEV           = 13.6058e0     # conversion from Ry to eV
EVTORY           = 1.0/13.6058e0   # conversion from eV to Ry
AMCONV           = 1.66042e-24/9.1095e-28*0.5e0  # mass conversion: a.m.u to a.u. (Ry)
UAKBAR           = 147105.e0  # pressure conversion from Ry/(a.u)^3 to K
DEGTORAD         = (3.14159265358979323846)/(180)  # Degrees to radians
CM1TOEV          = 1.23981e-4  # cm^-1 to eV
EVTOCM1          = 1.0/1.23981e-4  # eV to cm^-1

E2               = 2.0 # e^2

# Logos
PAO = 'PAO'
p = "\u03C0"
pp = p.encode('utf8')
PAOPY = str(PAO)+str(pp)

C = "\u00A9"
CC = C.encode('utf8')

LL = ['x','y','z']



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from mpi4py import MPI

from .do_atwfc_proj import *
from .write2xsf import write2xsf
from .communication import load_balancing

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_density ( data_controller, nr1, nr2, nr3 ):

  arry,attr = data_controller.data_dicts()

  # Calculation of the electron density

  if rank == 0 and attr['verbose']:
    print('Writing density files')

  rhoaux = np.zeros((nr1,nr2,nr3,attr['nspin']),dtype=complex,order="C")

  ini_ik,end_ik = load_balancing(comm.Get_size(), rank, attr['nkpnts'])
  
  basis = arry['basis']
  eps = 1.e-5
  for ispin in range(attr['nspin']):
    for ik in range(ini_ik,end_ik):
      gkspace = calc_gkspace(data_controller,ik,gamma_only=False)
      atwfcgk = calc_atwfc_k(basis,gkspace)
      oatwfcgk = ortho_atwfc_k(atwfcgk)
      atwfcr = fft_allwfc_G2R(oatwfcgk,gkspace, nr1, nr2, nr3, attr['omega'])
      for nb in range(attr['bnd']):
        if arry['E_k'][ik-ini_ik,nb,ispin] <= 0.0+eps:
          tmp = np.tensordot(arry['v_k'][ik-ini_ik,:,nb,ispin],atwfcr[:,:,:,:],axes=(0,0))
          rhoaux[:,:,:,ispin] += 2*np.conj(tmp)*tmp/attr['nkpnts']*attr['omega']/(nr1*nr2*nr3)

    rho = np.zeros((nr1,nr2,nr3,attr['nspin']),dtype=complex,order="C") if rank == 0 else None
    
    comm.Reduce(rhoaux,rho,op=MPI.SUM)
    rhoaux = None

    if rank == 0:
      fdensity = attr['outputdir']+'/density_%s.xsf'%str(ispin)
      write2xsf(data_controller,filename=fdensity,data=np.real(rho[:,:,:,ispin]))
  if rank == 0:
    if attr['verbose']: print('Total charge = ',np.real(np.sum(rho)).round(3))



# 
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_dielectric_tensor ( data_controller, ene ):
  from .constants import LL

  arrays,attributes = data_controller.data_dicts()

  smearing = attributes['smearing']
  if smearing != None and smearing != 'gauss' and smearing != 'm-p':
    if rank == 0:
      print('%s Smearing Not Implemented.'%smearing)
    quit()

  d_tensor = arrays['d_tensor']

  for ispin in range(attributes['nspin']):
    for n in range(d_tensor.shape[0]):
      ipol = d_tensor[n][0]
      jpol = d_tensor[n][1]

      epsi,epsr,eels,jdos,ieps = do_epsilon(data_controller, ene, ispin, ipol, jpol)

      # Write files
      indices = (LL[ipol], LL[jpol], ispin)
      for ep,es in [(epsi,'epsi'),(epsr,'epsr'),(eels,'eels'),(jdos,'jdos'),(ieps,'ieps')]:
        fn = '%s_%s%s_%d.dat'%((es,)+indices)
        data_controller.write_file_row_col(fn, ene, ep)

      if rank == 0:
        renorm = np.sqrt((2./np.pi)*(ene[3]-ene[2])*np.sum(epsi*ene))
        print(ipol,jpol,' plasmon frequency = ',renorm,' eV')
        print(' integration over JDOS = ', (ene[3]-ene[2])*np.sum(jdos))


def do_epsilon ( data_controller, ene, ispin, ipol, jpol ):
  from .constants import EPS0, EVTORY, RYTOEV

  # Compute the dielectric tensor

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  if ene[0] == 0.:
    ene[0] = .00001

  #=======================
  # EPS
  #=======================
  epsi_aux,epsr_aux,jdos_aux,count_aux = eps_loop(data_controller, ene, ispin, ipol, jpol)

  ### TNeeds revision. Each processor is allocating zeros here, when only rank 0 needs it. 
  ### Can be condensed
  epsi = np.zeros(esize, dtype=float)
  comm.Allreduce(epsi_aux, epsi, op=MPI.SUM)
  epsi_aux = None

  epsr = np.zeros(esize, dtype=float)
  comm.Allreduce(epsr_aux, epsr, op=MPI.SUM)
  epsr_aux = None

  epsr_aux = epsr_kramerskronig(data_controller, ene, epsi)
  epsr0 = np.zeros(esize, dtype=float)
  comm.Allreduce(epsr_aux, epsr0, op=MPI.SUM)
  epsr_aux = None

  jdos = np.zeros(esize, dtype=float)
  comm.Allreduce(jdos_aux, jdos, op=MPI.SUM)
  jods_aux = None

  count = np.zeros(1,dtype=float)
  comm.Allreduce(count_aux, count, op=MPI.SUM)
  count_aux = None

  ieps = np.zeros(esize, dtype=float)

  kq_wght = 1./attributes['nkpnts']
  epsi *= 64.0*np.pi*kq_wght/(attributes['omega'])
  # includes correction for apparent rigid shift of epsr - solved by getting the right e -> 0 limit from KK.
  if not attributes['metal']:
    epsr =  1. + epsr*64.0*np.pi/(attributes['omega']*attributes['nkpnts']) - (epsr[4]-epsr0[4])*64.0*np.pi/(attributes['omega']*attributes['nkpnts'])
  else:
    epsr =  1. + epsr*64.0*np.pi/(attributes['omega']*attributes['nkpnts']) 
  eels = epsi/(epsi**2+epsr**2)
  for i in range(esize):
    for j in range(1,esize):
        ieps[i] += ene[j]*epsi[j]/(ene[i]**2+ene[j]**2)
  ieps = 1.0 + (2./np.pi)*ieps*(ene[3]-ene[2])
  jdos /= (4.*count[0])

  return(epsi, epsr, eels, jdos, ieps)


def eps_loop ( data_controller, ene, ispin, ipol, jpol):
  from .constants import EPS0, EVTORY, RYTOEV, BOHR_RADIUS_ANGS
  from .smearing import intgaussian,gaussian,intmetpax,metpax

### What is this?
  orig_over_err = np.geterr()['over']
  np.seterr(over='raise')

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  bnd = attributes['bnd']
  temp = attributes['temp']
  delta = attributes['delta']
  snktot = arrays['pksp'].shape[0]
  smearing = attributes['smearing']

  Ef = 0.
  eps=1.e-8
  kq_wght = 1./attributes['nkpnts']

  jdos = np.zeros(esize, dtype=float)
  epsi = np.zeros(esize, dtype=float)
  epsr = np.zeros(esize, dtype=float)

  fn = None
  if smearing == None:
    fn = 2.*1./(1.+np.exp(arrays['E_k'][:,:bnd,ispin]/temp))
  elif smearing == 'gauss':
    fn = 2.*intgaussian(arrays['E_k'][:,:bnd,ispin], Ef, arrays['deltakp'][:,:bnd,ispin])
  elif smearing == 'm-p':
    fn = 2.*intmetpax(arrays['E_k'][:,:bnd,ispin], Ef, arrays['deltakp'][:,:bnd,ispin])

  # apparently there are numerical instabilities if energy levels are not completely occupied or completely empty - needs to be tested for metals
  if not attributes['metal']:
    fn = np.round(fn,0)
  count = np.zeros(1,dtype=float)

  bndmax = bnd

  for ik in range(fn.shape[0]):
    for iband2 in range(bndmax):
       for iband1 in range(bndmax):
          if iband1 != iband2:
             E_diff_nm = arrays['E_k'][ik,iband2,ispin] - arrays['E_k'][ik,iband1,ispin]
             f_nm =  fn[ik,iband2]-fn[ik,iband1]
             if np.abs(f_nm) > 2.e-3 and fn[ik,iband1] > 1.e-4 and fn[ik,iband2] < 2.0:
                pksp2 = np.real(arrays['pksp'][ik,ipol,iband1,iband2,ispin]*arrays['pksp'][ik,jpol,iband2,iband1,ispin])
                pksp2 *= attributes['alat']*BOHR_RADIUS_ANGS/(EPS0*RYTOEV)
                epsi[:] +=  pksp2*delta*ene[:]*fn[ik,iband1]/(((E_diff_nm**2-ene[:]**2)**2+delta**2*ene[:]**2)*(E_diff_nm))
                epsr[:] +=  pksp2*(E_diff_nm**2-ene[:]**2)*fn[ik,iband1]/(((E_diff_nm**2-ene[:]**2)**2+delta**2*ene[:]**2)*(E_diff_nm))
                jdos[:] +=  delta*(fn[ik,iband1]-fn[ik,iband2])/(np.pi*((E_diff_nm-ene[:])**2+delta**2))
                count[0] += (fn[ik,iband1]-fn[ik,iband2])

  if attributes['metal']:
    if rank == 0: print('NOT TESTED - needs different delta for intraband transitions and degauss from QE + check on units!!!')
    degauss=0.05
    fnF = None
    if smearing is None:
      fnF = np.empty((snktot,bnd), dtype=float)
      for n in range(bnd):
        for i in range(snktot):
          try:
            fnF[i,n] = .5/(1.+np.cosh(arrays['E_k'][i,n,ispin]/temp))
          except:
            fnF[i,n] = 1e8
      fnF /= temp
    elif smearing == 'gauss':
 ## Why .03* here?
      fnF = gaussian(arrays['E_k'][:,:bnd,ispin], Ef, .03*arrays['deltakp'][:,:bnd,ispin])
    elif smearing == 'm-p':
      fnF = metpax(arrays['E_k'][:,:bnd,ispin], Ef, arrays['deltakp'][:,:bnd,ispin])

    for ik in range(fn.shape[0]):
      for iband1 in range(bndmax):
        pksp2 = np.real(arrays['pksp'][ik,ipol,iband1,iband1,ispin]*arrays['pksp'][ik,jpol,iband1,iband1,ispin])
        pksp2 *= attributes['alat']*BOHR_RADIUS_ANGS/(EPS0*RYTOEV**3)
        epsi[:] +=  pksp2*delta*ene[:]*fnF[ik,iband1]/((ene[:]**4+delta**2*ene[:]**2)*degauss)
        epsr[:] -=  pksp2*fnF[ik,iband1]*ene[:]**2/((ene[:]**4+delta**2*ene[:]**2)*degauss)

  np.seterr(over=orig_over_err)
  return(epsi, epsr, jdos, count)


def epsr_kramerskronig ( data_controller, ene, epsi ):
  from .smearing import intmetpax
  from scipy.integrate import simps
  from .communication import load_balancing

  arrays,attributes = data_controller.data_dicts()

  esize = ene.size
  de = ene[1] - ene[0]

  epsr = np.zeros(esize, dtype=float)

  ini_ie,end_ie = load_balancing(comm.Get_size(), rank, esize)

  # Range checks for Simpson Integrals
  if end_ie == ini_ie:
    return
  if ini_ie < 3:
    ini_ie = 3
  if end_ie == esize:
    end_ie = esize-1

  f_ene = intmetpax(ene, attributes['shift'], 1.)
  for ie in range(ini_ie, end_ie):
    I1 = simps(ene[1:(ie-1)]*de*epsi[1:(ie-1)]*f_ene[1:(ie-1)]/(ene[1:(ie-1)]**2-ene[ie]**2))
    I2 = simps(ene[(ie+1):esize]*de*epsi[(ie+1):esize]*f_ene[(ie+1):esize]/(ene[(ie+1):esize]**2-ene[ie]**2))
    epsr[ie] = 2.*(I1+I2)/np.pi

  return epsr



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from mpi4py import MPI

comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def do_dos ( data_controller, emin, emax, ne, delta ):

  arry,attr = data_controller.data_dicts()
  bnd = attr['bnd']
  netot = attr['nkpnts']*bnd
  emax = np.amin(np.array([attr['shift'], emax]))
  arry['dos'] = np.empty((ne,), dtype=float)
  # DOS calculation with gaussian smearing
  ene = np.linspace(emin, emax, ne)

  if rank == 0 and attr['verbose']:
    print('Writing DoS Files')

  for ispin in range(attr['nspin']):

    dosaux = np.zeros((ne), order="C")

    E_k = arry['E_k'][:,:bnd,ispin]

    for n in range(ne):
      dosaux[n] = np.sum(np.exp(-((ene[n]-E_k)/delta)**2))

    dos = np.zeros((ne), dtype=float) if rank == 0 else None

    comm.Reduce(dosaux,dos,op=MPI.SUM)
    dosaux = None

    if rank == 0:
      dos *= float(bnd)/(float(netot)*np.sqrt(np.pi)*delta)
      arry['dos'] = dos
    fdos = 'dos_%s.dat'%str(ispin)
    data_controller.write_file_row_col(fdos, ene, dos)
    data_controller.broadcast_single_array('dos', dtype=float)
    #return dos if rank==0 else None

def do_dos_adaptive ( data_controller, emin, emax, ne ):
  from .smearing import gaussian, metpax

  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  arry,attr = data_controller.data_dicts()

  # DOS calculation with adaptive smearing
  ene = np.linspace(emin, emax, ne)
  arry['dosdk'] = np.empty((ne,), dtype=float)

  bnd = attr['bnd']
  netot = attr['nkpnts']*bnd

  if rank == 0 and attr['verbose']:
    print('Writing Adaptive DoS Files')

  for ispin in range(attr['nspin']):

    E_k = arry['E_k'][:,:bnd,ispin].reshape(arry['E_k'].shape[0]*bnd)
    delta = np.ravel(arry['deltakp'][:,:bnd,ispin], order='C')

    dosaux = np.zeros((ne), dtype=float)

    for n in range(ne):
      if attr['smearing'] == 'gauss':
        # adaptive Gaussian smearing
        dosaux[n] = np.sum(gaussian(ene[n],E_k,delta))

      elif attr['smearing'] == 'm-p':
        # adaptive Methfessel and Paxton smearing
        dosaux[n] = np.sum(metpax(ene[n],E_k,delta))

    dos = np.zeros((ne), dtype=float) if rank==0 else None
    comm.Reduce(dosaux, dos, op=MPI.SUM)
    dosaux = None

    if rank == 0:
      dos *= float(bnd)/netot
      arry['dosdk'] = dos
    fdosdk = 'dosdk_%s.dat'%str(ispin)
    data_controller.write_file_row_col(fdosdk, ene,dos)
    data_controller.broadcast_single_array('dosdk', dtype=float)




#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#
#
# References:
# Luis A. Agapito, Andrea Ferretti, Arrigo Calzolari, Stefano Curtarolo and Marco Buongiorno Nardelli,
# Effective and accurate representation of extended Bloch states on finite Hilbert spaces, Phys. Rev. B 88, 165127 (2013).
#
# Luis A. Agapito, Sohrab Ismail-Beigi, Stefano Curtarolo, Marco Fornari and Marco Buongiorno Nardelli,
# Accurate Tight-Binding Hamiltonian Matrices from Ab-Initio Calculations: Minimal Basis Sets, Phys. Rev. B 93, 035104 (2016).
#
# Luis A. Agapito, Marco Fornari, Davide Ceresoli, Andrea Ferretti, Stefano Curtarolo and Marco Buongiorno Nardelli,
# Accurate Tight-Binding Hamiltonians for 2D and Layered Materials, Phys. Rev. B 93, 125137 (2016).
#
# Pino D'Amico, Luis Agapito, Alessandra Catellani, Alice Ruini, Stefano Curtarolo, Marco Fornari, Marco Buongiorno Nardelli, 
# and Arrigo Calzolari, Accurate ab initio tight-binding Hamiltonians: Effective tools for electronic transport and 
# optical spectroscopy from first principles, Phys. Rev. B 94 165166 (2016).
# 

import numpy as np
import sys, time
from numpy import linalg as LAN

from mpi4py import MPI
from mpi4py.MPI import ANY_SOURCE
#from .load_balancing import *
from .get_R_grid_fft import *
from .communication import *
from .constants import *
from .perturb_split import *
# initialize parallel execution
comm=MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

try:
    from cuda_fft import *
except: pass
from scipy import fftpack as FFT


def do_d2Hd2k_ij(Hksp,Rfft,alat,npool,v_kp,bnd,degen):
    #----------------------
    # Compute the gradient of the k-space Hamiltonian
    #----------------------
    Rfft=np.transpose(Rfft,(3,0,1,2))

    num_n,nk1,nk2,nk3,nspin = Hksp.shape

    _,nk1,nk2,nk3,nspin = Hksp.shape
    
    M_ij   = np.zeros((6,v_kp.shape[0],bnd,v_kp.shape[3]),dtype=float,order="C")
    ij_ind = np.array([[0,0],[1,1],[2,2],[0,1],[0,2],[1,2]],dtype=int)

    nktot = nk1*nk2*nk3



    comm.Barrier()
    ########################################
    ### real space grid replaces k space ###
    ########################################
    # c1=c2=0
    # for ik in range(len(degen[0])):
    #     if len(degen[0][ik]) != 0:
    #         c1+=1
    #     else:
    #         c2+=1
    # print(c1+c2,c1,c2)
    

    #############################################################################################
    #############################################################################################
    #############################################################################################
    num_n = Hksp.shape[0]

    dvec_list=[]

    for ij in range(M_ij.shape[0]):
        dir_tmp=[]
        d2Hksp = None
        d2Hksp = np.zeros((num_n,nk1,nk2,nk3,nspin),dtype=complex,order='C')    
        
        ipol = ij_ind[ij][0]
        jpol = ij_ind[ij][1]

        RIJ = Rfft[ipol]*Rfft[jpol]

        for ispin in range(d2Hksp.shape[4]):
            for n in range(d2Hksp.shape[0]):                
                # because of the way this is coded...Hksp is actually HR*1.0j*alat
                d2Hksp[n,:,:,:,ispin] = FFT.fftn(RIJ*Hksp[n,:,:,:,ispin]*1.0j*alat)

        #############################################################################################
        #############################################################################################
        #############################################################################################

        #gather the arrays into flattened dHk
        d2Hksp = np.reshape(d2Hksp,(num_n,nk1*nk2*nk3,nspin),order='C')        
        d2Hksp = gather_scatter(d2Hksp,1,npool)
        nawf   = int(np.sqrt(d2Hksp.shape[0]))

        d2Hksp = np.reshape(d2Hksp,(nawf,nawf,d2Hksp.shape[1],nspin),order='C')

        tksp = np.zeros_like(d2Hksp)

        #find non-degenerate set of psi(k) for d2H/d2k_ij
        for ispin in range(tksp.shape[3]):
            isp_tmp=[]
            for ik in range(tksp.shape[2]):

                # we save dvec so that it can be used when calculating the second term in d2E/d2k
                tksp[:,:,ik,ispin],_,dvec = perturb_split(d2Hksp[:,:,ik,ispin],
                                                          d2Hksp[:,:,ik,ispin],
                                                          v_kp[ik,:,:,ispin],
                                                          degen[ispin][ik],return_v_k=True)

                isp_tmp.append(dvec)
            dir_tmp.append(isp_tmp)
        dvec_list.append(dir_tmp)

        
        # get the value for d2H/d2k
        for ispin in range(d2Hksp.shape[3]):
            for n in range(bnd):                
                M_ij[ij,:,n,ispin] = tksp[n,n,:,ispin].real

        comm.Barrier()

    Hksp_aux= d2Hksp = None    

    return M_ij,dvec_list






#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#
import numpy as np
import sys, time
import multiprocessing

import pycuda.autoinit
import pycuda.gpuarray as gpuarray
import skcuda.fft as skfft


# Perform an inverse FFT on 'axes' of 'Hk'
# Restriction: len(Hk) >= len(axes)
# Restriction: 'axes' must be a list of unique, monotonically increasing, integers
#               beginning with 0.
def cuda_ifftn ( Hk, axes=[0,1,2] ):
    return cuda_efftn(Hk, axes, False)

# Perform an FFT on 'axes' of 'Hr'
# Restriction: len(Hr) >= len(axes)
# Restriction: 'axes' must be a list of unique, monotonically increasing, integers
#               beginning with 0.
def cuda_fftn ( Hr, axes=[0,1,2] ):
    return cuda_efftn(Hr, axes, True)


def cuda_efftn(H, axes, forward):
    hShape = H.shape
    hDim = len(hShape)
    fftDim = len(axes)

    # Reshape 'axes' to be the array's end dimensions and ensure contiguity
    H = np.ascontiguousarray(np.moveaxis(H, axes, np.arange(hDim-fftDim, hDim, 1)))

    # Calculate number of batches
    batchSize = 1
    for i in range(hDim-fftDim):
        batchSize *= H.shape[i]

    # Reshape to accomodate batching
    H = np.reshape(H, (batchSize, H.shape[hDim-3], H.shape[hDim-2], H.shape[hDim-1]))

    # Pass array to the GPU and perform iFFT on each batch
    H_gpu = gpuarray.to_gpu(H)
    plan = skfft.Plan(H_gpu.shape[1:fftDim+1], H.dtype, H.dtype, H_gpu.shape[0])

    if forward:
        skfft.fft(H_gpu, H_gpu, plan)
    else:
        skfft.ifft(H_gpu, H_gpu, plan, True)

    # Reshape to original dimensions
    H = np.moveaxis(H_gpu.get(), 0, fftDim)
    H = np.reshape(H, hShape)

    return H



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#
#
# References:
# Luis A. Agapito, Andrea Ferretti, Arrigo Calzolari, Stefano Curtarolo and Marco Buongiorno Nardelli,
# Effective and accurate representation of extended Bloch states on finite Hilbert spaces, Phys. Rev. B 88, 165127 (2013).
#
# Luis A. Agapito, Sohrab Ismail-Beigi, Stefano Curtarolo, Marco Fornari and Marco Buongiorno Nardelli,
# Accurate Tight-Binding Hamiltonian Matrices from Ab-Initio Calculations: Minimal Basis Sets, Phys. Rev. B 93, 035104 (2016).
#
# Luis A. Agapito, Marco Fornari, Davide Ceresoli, Andrea Ferretti, Stefano Curtarolo and Marco Buongiorno Nardelli,
# Accurate Tight-Binding Hamiltonians for 2D and Layered Materials, Phys. Rev. B 93, 125137 (2016).
#
# Pino D'Amico, Luis Agapito, Alessandra Catellani, Alice Ruini, Stefano Curtarolo, Marco Fornari, Marco Buongiorno Nardelli, 
# and Arrigo Calzolari, Accurate ab initio tight-binding Hamiltonians: Effective tools for electronic transport and 
# optical spectroscopy from first principles, Phys. Rev. B 94 165166 (2016).
# 

def do_gradient ( data_controller ):
  import numpy as np
  from scipy import fftpack as FFT
  from .get_R_grid_fft import get_R_grid_fft

  arry,attr = data_controller.data_dicts()

  #----------------------
  # Compute the gradient of the k-space Hamiltonian
  #----------------------

  nktot = attr['nkpnts']
  snawf,nk1,nk2,nk3,nspin = arry['Hksp'].shape

  # fft grid in R shifted to have (0,0,0) in the center
  get_R_grid_fft(data_controller, nk1, nk2, nk3)

  arry['dHksp'] = np.empty((snawf,nk1,nk2,nk3,3,nspin), dtype=complex, order='C')
  for ispin in range(nspin):
    for n in range(snawf):
      ########################################
      ### real space grid replaces k space ###
      ########################################
      if attr['use_cuda']:
        arry['Hksp'][n,:,:,:,ispin] = cuda_ifftn(arry['Hksp'][n,:,:,:,ispin])*1.0j*attr['alat']
      else:
        arry['Hksp'][n,:,:,:,ispin] = FFT.ifftn(arry['Hksp'][n,:,:,:,ispin])*1.0j*attr['alat']

      # Compute R*H(R)
      for l in range(3):
        arry['dHksp'][n,:,:,:,l,ispin] = FFT.fftn(arry['Rfft'][:,:,:,l]*arry['Hksp'][n,:,:,:,ispin])




# 
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


def gaussian ( eig, ene, delta ):
    import numpy as np

    # Gaussian Smearing
    return (np.exp(-((ene-eig)/delta)**2)/delta)/np.sqrt(np.pi)



def metpax ( eig, ene, delta ):
    import numpy as np
    from math import factorial
    from numpy.polynomial.hermite import hermval

    # Methfessel and Paxton smearing
    nh = 5
    coeff = np.zeros(2*nh)
    coeff[0] = 1.
    for n in range(2,2*nh,2):
        m = n/2
        coeff[n] = (-1.)**m/(factorial(m)*(4.0**m)*np.sqrt(np.pi))

    x = (ene-eig)/delta
    return hermval(x, coeff)*np.exp(-(x)**2)/(delta*np.sqrt(np.pi))

def intgaussian ( eig, ene, delta ):
    from scipy.special import erf

    # integral of the gaussian function as approximation of the Fermi-Dirac distribution 
    return (1.-erf((eig-ene)/delta))/2.

def intmetpax ( eig, ene, delta ):
    import numpy as np
    from math import factorial
    from scipy.special import erf
    from numpy.polynomial.hermite import hermval

    # Methfessel and Paxton correction to the Fermi-Dirac distribution
    nh = 5
    coeff = np.zeros(2*nh)
    coeff[0] = 0.
    for n in range(2,2*nh,2):
        m = n/2
        coeff[n-1] = (-1.)**m/(factorial(m)*(4.0**m)*np.sqrt(np.pi))

    x = (eig-ene)/delta
    return (1.-erf(x))/2. + hermval(x, coeff)*np.exp(-(x**2))/np.sqrt(np.pi)



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016,2017 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#
import numpy as np

def zero_pad(aux,nk1,nk2,nk3,nfft1,nfft2,nfft3):
    '''
    Pad frequency domain with zeroes, such that any relationship between
        aux[k] and aux[N-k] is preserved.

    Arguments:
        aux (ndarray): unpadded frequency domain data
        nk1 (int): current size of aux along axis 0
        nk2 (int): current size of aux along axis 1
        nk3 (int): current size of aux along axis 2
        nfft1 (int): number of zeroes to pad axis 0 by
        nfft1 (int): number of zeroes to pad axis 1 by
        nfft1 (int): number of zeroes to pad axis 2 by

    Returns:
        auxp3 (ndarray): padded frequency domain data
    '''
    # post-padding dimensions
    nk1p = nfft1+nk1
    nk2p = nfft2+nk2
    nk3p = nfft3+nk3
    # halfway points
    sk1 = int((nk1+1)/2)
    sk2 = int((nk2+1)/2)
    sk3 = int((nk3+1)/2)
    # parities (even <-> p==1)
    p1 = (nk1 & 1)^1
    p2 = (nk2 & 1)^1
    p3 = (nk3 & 1)^1

    # accomodate nfft==0
    if nfft1 == 0:  p1 = 0
    if nfft2 == 0:  p2 = 0
    if nfft3 == 0:  p3 = 0

    # first dimension
    auxp1 = np.zeros((nk1,nk2,nk3p),dtype=complex)
    auxp1[:,:,:sk3+p3]=aux[:,:,:sk3+p3]
    auxp1[:,:,nfft3+sk3:]=aux[:,:,sk3:]
    # second dimension
    auxp2 = np.zeros((nk1,nk2p,nk3p),dtype=complex)
    auxp2[:,:sk2+p2,:]=auxp1[:,:sk2+p2,:]
    auxp2[:,nfft2+sk2:,:]=auxp1[:,sk2:,:]
    # third dimension
    auxp3 = np.zeros((nk1p,nk2p,nk3p),dtype=complex)
    auxp3[:sk1+p1,:,:]=auxp2[:sk1+p1,:,:]
    auxp3[nfft1+sk1:,:,:]=auxp2[sk1:,:,:]

    # halve Nyquist axes
    if p1:
        auxp3[ sk1,:,:] /= 2
        auxp3[-sk1,:,:] /= 2
    if p2:
        auxp3[:, sk2,:] /= 2
        auxp3[:,-sk2,:] /= 2
    if p3:
        auxp3[:,:, sk3] /= 2
        auxp3[:,:,-sk3] /= 2

    return(auxp3)


def zero_pad_float(aux,nk1,nk2,nk3,nfft1,nfft2,nfft3):
    """ Deprecated. Use zero_pad instead.

    Note that this function uses the old padding algorithm, which
        1) does not (quite) preserve symmetry of DFT for even nk
        2) puts the zeros in the wrong spot altogether for odd nk...
    Besides that, it only works with real numbers...
    """
    # zero padding for FFT interpolation in 3D
    nk1p = nfft1+nk1
    nk2p = nfft2+nk2
    nk3p = nfft3+nk3
    # first dimension
    auxp1 = np.zeros((nk1,nk2,nk3p),dtype=float)
    auxp1[:,:,:int(nk3/2)]=aux[:,:,:int(nk3/2)]
    auxp1[:,:,int(nfft3+nk3/2):]=aux[:,:,int(nk3/2):]
    # second dimension
    auxp2 = np.zeros((nk1,nk2p,nk3p),dtype=float)
    auxp2[:,:int(nk2/2),:]=auxp1[:,:int(nk2/2),:]
    auxp2[:,int(nfft2+nk2/2):,:]=auxp1[:,int(nk2/2):,:]
    # third dimension
    auxp3 = np.zeros((nk1p,nk2p,nk3p),dtype=float)
    auxp3[:int(nk1/2),:,:]=auxp2[:int(nk1/2),:,:]
    auxp3[int(nfft1+nk1/2):,:,:]=auxp2[int(nk1/2):,:,:]

    return(auxp3)



#
# PAOpy
#
# Utility to construct and operate on Hamliltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016 ERMES group (http://ermes.unt.edu)
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#
#
# References:
# Luis A. Agapito, Andrea Ferretti, Arrigo Calzolari, Stefano Curtarolo and Marco Buongiorno Nardelli,
# Effective and accurate representation of extended Bloch states on finite Hilbert spaces, Phys. Rev. B 88, 165127 (2013).
#
# Luis A. Agapito, Sohrab Ismail-Beigi, Stefano Curtarolo, Marco Fornari and Marco Buongiorno Nardelli,
# Accurate Tight-Binding Hamiltonian Matrices from Ab-Initio Calculations: Minimal Basis Sets, Phys. Rev. B 93, 035104 (2016).
#
# Luis A. Agapito, Marco Fornari, Davide Ceresoli, Andrea Ferretti, Stefano Curtarolo and Marco Buongiorno Nardelli,
# Accurate Tight-Binding Hamiltonians for 2D and Layered Materials, Phys. Rev. B 93, 125137 (2016).
#
# Pino D'Amico, Luis gapito, Alessandra Catellani, Alice Ruini, Stefano Curtarolo, Marco Fornari, Marco Buongiorno Nardelli, 
# and Arrigo Calzolari, Accurate ab initio tight-binding Hamiltonians: Effective tools for electronic transport and 
# optical spectroscopy from first principles, Phys. Rev. B 94 165166 (2016).
# 


import scipy.optimize as so
#import matplotlib.pyplot as plt
from scipy import fftpack as FFT
import numpy as np
import cmath
import sys
import scipy
from scipy import fftpack as FFT
from mpi4py import MPI
from mpi4py.MPI import ANY_SOURCE
import os
import scipy.optimize as OP
from numpy import linalg as LAN
from .communication import gather_full, scatter_full
from numpy import linalg as LAN

# initialize parallel execution
comm=MPI.COMM_WORLD
rank = comm.Get_rank()
size = comm.Get_size()

#np.set_printoptions(precision=8, threshold=100, edgeitems=50, linewidth=350, suppress=True)

def band_loop_H ( ini_ik, end_ik, HRaux, kq, R ):
  nawf,_,_,nspin = HRaux.shape
  kdot = np.zeros((1,R.shape[0]), dtype=complex, order='C')
  kdot = np.tensordot(R, 2.0j*np.pi*kq[:,ini_ik:end_ik], axes=([1],[0]))
  np.exp(kdot, kdot)

  auxh = np.zeros((nawf,nawf,1,nspin), dtype=complex, order='C')
  for ispin in range(nspin):
    auxh[:,:,ini_ik:end_ik,ispin] = np.tensordot(HRaux[:,:,:,ispin], kdot, axes=([2],[0]))
  return auxh

def gen_eigs ( HRaux, kq, R ):
  # Load balancing

  nawf,_,_,nspin = HRaux.shape
  E_kp = np.zeros((1,nawf,nspin), dtype=np.float64)

  kq=kq[:,None]

#  Hks_int  = np.zeros((nawf,nawf,1,nspin),dtype=complex,order='C') # final data arrays
  Hks_int = band_loop_H(0,1,HRaux,kq,R)

  for ispin in range(nspin):
    E_kp[:,:,ispin] =  LAN.eigvalsh(Hks_int[:,:,0,ispin],UPLO='U')

  return E_kp


def get_gap ( HR, kq, R, nelec ):
  E_kp = gen_eigs(HR, kq, R)
  return E_kp[0,nelec,0] - E_kp[0,nelec-1,0]


def get_R_grid_fft ( nr1, nr2, nr3 ):

  R = np.zeros((nr1*nr2*nr3,3))

  for i in range(nr1):
    for j in range(nr2):
      for k in range(nr3):
        n = k + j*nr3 + i*nr2*nr3
        Rx = float(i)/float(nr1)
        Ry = float(j)/float(nr2)
        Rz = float(k)/float(nr3)
        if Rx >= 0.5: Rx=Rx-1.0
        if Ry >= 0.5: Ry=Ry-1.0
        if Rz >= 0.5: Rz=Rz-1.0
        Rx -= int(Rx)
        Ry -= int(Ry)
        Rz -= int(Rz)

        R[n,0] = Rx*nr1 
        R[n,1] = Ry*nr2 
        R[n,2] = Rz*nr3

  return R  


def get_search_grid ( nk1, nk2, nk3, snk1_range=[-0.5,0.5], snk2_range=[-0.5,0.5], snk3_range=[-0.5,0.5], endpoint=False ):

  nk1_arr = np.linspace(snk1_range[0], snk1_range[1], num=nk1, endpoint=endpoint)
  nk2_arr = np.linspace(snk2_range[0], snk2_range[1], num=nk2, endpoint=endpoint)
  nk3_arr = np.linspace(snk3_range[0], snk3_range[1], num=nk3, endpoint=endpoint)

  #nk_str = np.zeros((nk1*nk2*nk3,3), order='C')
  return np.array(np.meshgrid(nk1_arr,nk2_arr,nk3_arr,indexing='ij')).T.reshape(-1,3)

def find_weyl ( data_controller, test_rad, search_grid ):

  arry,attr = data_controller.data_dicts()

  symf = attr['symmetrize']
  nelec,verbose = attr['nelec'],attr['verbose']
  HRs,symops,b_vectors,TR_flag = arry['HRs'],arry['sym_rot'],arry['b_vectors'],arry['sym_TR']

  nawf,_,nk1,nk2,nk3,nspin = HRs.shape
  R = get_R_grid_fft(nk1, nk2, nk3)

  HRs = np.reshape(HRs, (nawf,nawf,nk1*nk2*nk3,nspin))

  mag_soc = np.logical_and(attr["dftMAG"], attr["dftSO"])

  CAND,ene = find_min(HRs, nelec, R, b_vectors, symf, verbose, search_grid)
  
  WEYL = {}
  if rank == 0:
    if symf:
      # get all equiv k
      CAND = get_equiv_k(CAND, symops, TR_flag, mag_soc)

    if verbose:
      for i in range(CAND.shape[0]):
        tup = (i+1,) + (CAND[i][j] for j in range(3))
        print('Weyl point candidate #%d crystal coord: [ % 5.4f % 5.4f % 5.4f ]\n'%tup)
      for i in range(CAND.shape[0]):
        in_cart = b_vectors.T.dot(CAND[i])
        tup = (i+1,) + (in_cart[j] for j in range(3))
        print('Weyl point candidate #%d 2pi/alat: [ % 5.4f % 5.4f % 5.4f ]'%tup)

    try:
      import z2pack
      import tbmodels      
      import os
      model = tbmodels.Model.from_wannier_files(hr_file=os.path.join(attr['opath'],'z2pack_hamiltonian.dat'))
      system = z2pack.tb.System(model,bands=nelec)

      candidates=0

      if not verbose:
        import logging
        logging.getLogger('z2pack').setLevel(logging.WARNING)

      for kq in CAND:
        # if distance between two candidates is very small
        k_rad = np.amin(np.sqrt(np.sum((kq-CAND)**2,axis=1)))*0.5
        if k_rad > test_rad:
          k_rad = test_rad
        k_rad = 0.005

        surface = z2pack.shape.Sphere(center=tuple(kq), radius=k_rad)
        result_1 = z2pack.surface.run(system=system, surface=surface)
        invariant = z2pack.invariant.chern(result_1)

        if invariant != 0:
          candidates += 1
          WEYL[str(kq).replace(',','')] = invariant

    except:
      print('Could not load z2pack to verify chirality of weyl points')
      for kq in CAND:
        WEYL[str(kq).replace(',', '')] = '?'

    if verbose:
      print()

    wcs = '{0:>3} {1:>10} {2:>10} {3:>10} {4:>2}\n'.format('#','2pi/a','2pi/b','2pi/c','C')
    wcs += '-'*39 + '\n'

    for j,k in enumerate(WEYL.keys()):
      fm = np.array(list(map(float,k[1:-1].split())))     
      fm = b_vectors.T.dot(fm[:,None])[:,0]
      try:
        if verbose:
          fstring = 'Found Candidate No. {0} at [{1:>7.4f} {2:>7.4f} {3:>7.4f}] with Chirality:{4:>2}'
          print(fstring.format(j+1,fm[0],fm[1],fm[2],int(WEYL[k])))
        wcs += '{0:>3d} {1:>10.4f} {2:>10.4f} {3:>10.4f} {4:>2}\n'.format(j+1,fm[0],fm[1],fm[2],int(WEYL[k]))
      except:
        wcs += '{0:>3d} {1:>10.4f} {2:>10.4f} {3:>10.4f} {4:>2}\n'.format(j+1,fm[0],fm[1],fm[2],WEYL[k])

    with open(os.path.join(attr['opath'],'weyl_points.dat'), 'w') as ofo:
      ofo.write(wcs)

def find_min ( HRs, nelec, R, a_vectors, symf, verbose, search_grid=[8,8,8] ):

  snk = tuple(search_grid[i] for i in range(3))
  #snk2 = search_grid[1]
  #snk3 = search_grid[2]
  #search_grid = do_search_grid(snk1,snk2,snk3)
  search_grid = get_search_grid(*snk)
  
  #get the search grid off possible HSP
  end = np.array([.5]*3)
  start = np.array([-.5]*3)
  #end1 = end2 = end3 = 0.5
  #start1 = start2 = start3 = -0.5
  bounds_K  = np.zeros((search_grid.shape[0],3,2))
  #guess_K   = np.zeros((search_grid.shape[0],3))

  #do the bounds for each search subsection of FBZ
  #search in boxes
  bounds_K[:,:,0] = search_grid
  bbox = np.array([(end[i]-start[i])/snk[i] for i in range(3)])
  bounds_K[:,:,1] = search_grid + bbox
  #initial guess is in the middle of each box
  guess_K = search_grid + 0.5*bbox

  sgi = np.arange(bounds_K.shape[0], dtype=int)
  sgi = scatter_full(sgi, 1)

  candidates = np.zeros((sgi.shape[0],3))

  lam_XiP = lambda K: get_gap(HRs, K, R, nelec)
  for i in range(sgi.shape[0]):
    solx = OP.minimize(lam_XiP, guess_K[sgi[i]], bounds=bounds_K[sgi[i]], method="L-BFGS-B", options={"ftol":1.e-14,"gtol":1.e-12})

    if np.abs(solx.fun) < 1.e-5:
      candidates[i] = solx.x
      
  candidates = gather_full(candidates,1)

  ene = None
  if rank == 0:  
    # filter out non hits
    candidates = candidates[np.where(np.sum(candidates,axis=1)!=0.0)]
    # calculate energy at each candidate to reduce equiv ones
    ene = np.zeros((candidates.shape[0]))
    gaps = np.zeros((candidates.shape[0]))
    for i in range(ene.shape[0]):
      eigs = gen_eigs(HRs,candidates[i],R)[0,:,0]
      ene[i] = eigs[nelec]
      gaps[i] = eigs[nelec]-eigs[nelec-1]

    if symf:
      # sort by gap size (should be nearly zero)
      idx = np.argsort(gaps)
      ene = ene[idx]
      candidates = candidates[idx]     
      # filter out duplicates by degeneracy
      _,idx=np.unique(np.around(ene,decimals=4),return_index=True)
      ene = ene[idx]
      candidates = candidates[idx]
      if verbose:
       print('\nfound %s non-equivilent candidates'%candidates.shape[0])
       for i in range(ene.shape[0]):
         tup = (candidates[i][j] for j in range(3)) + (ene[i],)
         print("[ % 7.4f % 7.4f % 7.4f ] % 4.4f"%tup)
       print()

  comm.Barrier()
  return (candidates, ene)


def get_equiv_k(kp,symop,sym_TR,mag_soc):
  from .pao_sym import correct_roundoff

  # if we have time inversion sym
  if not mag_soc:
     kp=np.vstack([kp,-kp])
  

  kp = correct_roundoff(kp)
  newk_tot = np.copy(kp)

  for isym in range(symop.shape[0]):
    #transform k -> k' with the sym op
    if sym_TR[isym]:
      newk = ((((-symop[isym] @ (kp.T%1.0))%1.0)+0.5)%1.0)-0.5
    else:
      newk = (((( symop[isym] @ (kp.T%1.0))%1.0)+0.5)%1.0)-0.5

    newk = correct_roundoff(newk)
    newk[np.where(np.isclose(newk,0.5))]=-0.5
    newk[np.where(np.isclose(newk,-1.0))]=0.0
    newk[np.where(np.isclose(newk,1.0))]=0.0

    newk_tot=np.vstack([newk_tot,newk.T])

  # filter duplicates
  newk_r=np.around(newk_tot,decimals=4)
  newk_r[np.where(np.abs(newk_r)<1.e-4)]=0

  b_pos = np.ascontiguousarray(newk_r).view(np.dtype((np.void,newk_r.dtype.itemsize *newk_r.shape[1])))
  _, idx = np.unique(b_pos, return_index=True)
  newk_tot=newk_tot[idx]

  return newk_tot



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

import numpy as np
from .get_K_grid_fft import get_K_grid_fft_crystal

from mpi4py import MPI
comm = MPI.COMM_WORLD
rank = comm.Get_rank()

def write4bt2(data_controller):
	
	'''
		Write data in the GENE format suitable to be read by BoltzTrap2
		Returns also band derivatives (momentum matix) if available
		It shuld be called after:
		paoflow.pao_hamiltonian()
		paoflow.pao_eigh()
		paoflow.gradient_and_momenta()
		optional: call paoflow.interpolated_hamiltonian() to improve on the initial meshh of k-points
		
	'''
	
	if rank == 0:
		
		Ry2eV = 13.60569193
		arry,attr = data_controller.data_dicts()
		
		arry['kgrid'] = get_K_grid_fft_crystal(attr['nk1'],attr['nk2'],attr['nk3'])
		
		# write data in the GENE format for BoltzTrap2
		prefix = attr['savedir'].split('.')[0]
		fname_energy = prefix + '.energy'
		fname_struct = prefix + '.structure'
		
		f_energy = prefix + '\n'
		f_energy += str(attr['nkpnts']) + ' ' + str(int(attr['nspin'])) + ' ' + str(0) + '\n'
		for ik in range(attr['nkpnts']):
			f_energy += str(arry['kgrid'][ik][0]) + ' ' +\
						str(arry['kgrid'][ik][1]) + ' ' +\
						str(arry['kgrid'][ik][2]) +' ' + str(attr['nbnds']) + '\n'
			for ib in range(attr['nbnds']):
				f_energy += str(arry['E_k'][ik,ib,0]/Ry2eV) + '\n'
		
		f = open(fname_energy, 'w')
		f.write(f_energy)
		f.close()
		
		f_struct = prefix + '\n'
		for i in range(3):
			f_struct += str(arry['a_vectors'][i][0]*attr['alat']) + ' ' +\
						str(arry['a_vectors'][i][1]*attr['alat']) + ' ' +\
						str(arry['a_vectors'][i][2]*attr['alat']) + '\n'
		f_struct += str(attr['natoms']) + '\n'
		for ia in range(attr['natoms']):
			f_struct += str(arry['atoms'][ia]) + ' '
			f_struct += str(arry['tau'][ia].dot(arry['b_vectors'].T)[0]/attr['alat']) + ' ' +\
						str(arry['tau'][ia].dot(arry['b_vectors'].T)[1]/attr['alat']) + ' ' +\
						str(arry['tau'][ia].dot(arry['b_vectors'].T)[2]/attr['alat']) + '\n'
		
		f = open(fname_struct, 'w')
		f.write(f_struct)
		f.close()
		
		try:
			mommat = np.zeros((attr['nkpnts'],attr['nbnds'],3),dtype=float)
			for ib in range(attr['nbnds']):
				mommat[:,ib,:] = -np.real(arry['pksp'][:,:,ib,ib,0])/(2*Ry2eV)
			return(mommat)
		except:
			print('momentum matrix not available')
			return(None)


#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


def do_fermisurf ( data_controller ):
  import numpy as np
  from mpi4py import MPI
  from os.path import join
  from .communication import gather_full

  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  arry,attr = data_controller.data_dicts()

  #maximum number of bands crossing fermi surface
  ###### PARALLELIZATION
  E_kf = gather_full(arry['E_k'], attr['npool'])

  if rank == 0:
    if attr['verbose']:
      print('Writing bxsf file for Fermi Surface')

    nawf,nktot = attr['nawf'],attr['nkpnts']
    nk1,nk2,nk3 = attr['nk1'],attr['nk2'],attr['nk3']
    fermi_up,fermi_dw = attr['fermi_up'],attr['fermi_dw']

    E_ks = np.reshape(E_kf, (nk1,nk2,nk3,nawf,attr['nspin']))

    for ispin in range(attr['nspin']):

      ind_plot = []
      eigband = []

      for ib in range(nawf):
        E_k_min = np.amin(E_kf[:,ib,ispin])
        E_k_max = np.amax(E_kf[:,ib,ispin])
        btwUp = E_k_min < fermi_up and E_k_max > fermi_up
        btwDwn = E_k_min < fermi_dw and E_k_max > fermi_dw
        btwUaD = E_k_min > fermi_dw and E_k_max < fermi_up
        if btwUp or btwDwn or btwUaD:
          ind_plot.append(ib)
          eigband.append(E_ks[:,:,:,ib,ispin])

      feig = 'FermiSurf_%d.bxsf'%ispin
      eigband = np.array(eigband)
      data_controller.write_bxsf(feig, np.moveaxis(eigband,0,3), len(ind_plot), indices=ind_plot)

      for i,ib in enumerate(eigband):
        np.savez(join(attr['opath'],'Fermi_surf_band_%d_%d'%(ind_plot[i],ispin)), nameband=ib)

  comm.Barrier()
  E_kf = E_ks = None



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#


def do_effective_mass( data_controller,ene ):

  from mpi4py import MPI
  import numpy as np
  from os.path import join
  from numpy import linalg as npl
  from .communication import gather_full
  from .get_K_grid_fft import get_K_grid_fft_crystal
  from os.path import join

  comm = MPI.COMM_WORLD
  rank = comm.Get_rank()

  ary,attr = data_controller.data_dicts()

  omega = attr['omega']
  bnd   = attr['bnd']
  nspin = attr['nspin']
  spin_mult = 1. if nspin==2 or attr['dftSO'] else 2.
  E_k   = ary['E_k']
  d2Ed2k= ary['d2Ed2k']
  kq_wght= ary['kq_wght']
  me = 9.1093837015e-31
  eminBT =np.amin(ene)
  emaxBT =np.amax(ene)

  en_buff=1.0

  #only count the states from emin-en_buff emin+en_buff
  E_k_mask = np.where(np.logical_and(E_k[:,:bnd,:]>=(eminBT-en_buff),E_k[:,:bnd,:]<=(emaxBT+en_buff)))
  E_k_range = np.ascontiguousarray(E_k[E_k_mask[0],E_k_mask[1],E_k_mask[2]])
  d2Ed2k_range = np.ascontiguousarray(d2Ed2k[:,E_k_mask[0],E_k_mask[1],E_k_mask[2]])

  write_masses=True
  if write_masses:
    # writing the effective masses to file
    #scale factor
    SI_conv = 0.036749302892341
    em_flat = d2Ed2k*SI_conv
    em_flat = np.ascontiguousarray(np.transpose(em_flat,axes=(1,2,3,0)))
    em_flat = gather_full(em_flat,attr['npool'])

    if rank==0:
      nk,bnd,nspin,_ = em_flat.shape
      em_flat = np.ascontiguousarray(np.transpose(em_flat,axes=(2,0,1,3)))

      em_tens=np.zeros((nspin,nk,bnd,3,3))
      e_mass =np.zeros((nspin,nk,bnd,8))

      # build the effective mass tensors from the flattened version
      em_tens[...,0,0]=em_flat[...,0]
      em_tens[...,1,1]=em_flat[...,1]
      em_tens[...,2,2]=em_flat[...,2]
      em_tens[...,0,1]=em_flat[...,3]
      em_tens[...,1,0]=em_flat[...,3]
      em_tens[...,0,2]=em_flat[...,4]
      em_tens[...,2,0]=em_flat[...,4]
      em_tens[...,1,2]=em_flat[...,5]
      em_tens[...,2,1]=em_flat[...,5]
      # diagonalize
      for sp in range(nspin):
        for k in range(nk):
          for b in range(bnd):
            effm =  np.linalg.eigvals(np.linalg.inv(em_tens[sp,k,b]))
            e_mass[sp,k,b,[4,5,6]] = effm

            if np.prod(effm)<0:
              dos_em = -np.prod(np.abs(effm))**(1.0/3.0)
            else:
              dos_em =  np.prod(np.abs(effm))**(1.0/3.0)

            e_mass[sp,k,b,7] = dos_em

    effm=dos_em=em_tens=em_flat=None

    E_k_temp=gather_full(E_k,attr['npool'])
    if rank==0:
      E_k_temp = np.transpose(E_k_temp,axes=(2,0,1))
      e_mass[...,3]  = E_k_temp[:,:,:attr['bnd']]
      e_mass[...,:3] = get_K_grid_fft_crystal(attr['nk1'],attr['nk2'],attr['nk3'])[None,:,None]

      for sp in range(nspin):
        fpath = join(attr['opath'],'effective_masses_%d.dat'%sp)
        with open(fpath,'w') as ofo:
          ofo.write('    k_1     k_2     k_3     E-E_f              m_1              m_2              m_3            m_dos\n')
          ofo.write('-'*101)
          ofo.write('\n')

          for sp in range(nspin):
            for k in range(nk):
              for b in range(bnd):
                ofo.write('% 4.4f % 4.4f % 4.4f % 9.4f % 16.4f % 16.4f % 16.4f % 16.4f\n'%tuple(e_mass[sp,k,b].tolist()))

    E_k_temp=None



#
# PAOFLOW
#
# Utility to construct and operate on Hamiltonians from the Projections of DFT wfc on Atomic Orbital bases (PAO)
#
# Copyright (C) 2016-2018 ERMES group (http://ermes.unt.edu, mbn@unt.edu)
#
# Reference:
# M. Buongiorno Nardelli, F. T. Cerasoli, M. Costa, S Curtarolo,R. De Gennaro, M. Fornari, L. Liyanage, A. Supka and H. Wang,
# PAOFLOW: A utility to construct and operate on ab initio Hamiltonians from the Projections of electronic wavefunctions on
# Atomic Orbital bases, including characterization of topological materials, Comp. Mat. Sci. vol. 143, 462 (2018).
#
# This file is distributed under the terms of the
# GNU General Public License. See the file `License'
# in the root directory of the present distribution,
# or http://www.gnu.org/copyleft/gpl.txt .
#

def spinor ( l, j, m, spin ):
  # This function calculates the numerical Clebsch-Gordan coefficients of a spinor
  # with orbital angular momentum l, total angular momentum j, 
  # projection along z of the total angular momentum m+-1/2. Spin selects
  # the up (spin=0) or down (spin=1) coefficient.

  import numpy as np
  from mpi4py import MPI

  rank = MPI.COMM_WORLD.Get_rank()

  if spin != 0 and spin != 1:
    if rank == 0:
      print('Spinor - spin direction unknown')
    quit()
  if m < -l-1 or m > l:
    if rank == 0:
      print('Spinor - m not allowed')
    quit()

  denom=1./(2.*l+1.)
  if (abs(j-l-.5) < 1.e-8):
      if spin == 0: spinor = np.sqrt((l+m+1.)*denom)
      if spin == 1: spinor = np.sqrt((l-m)*denom)
  elif abs(j-l+.5) < 1.e-8:
      if m < -l+1:
          spinor=0.
      else:
          if spin == 0: spinor = np.sqrt((l-m+1.)*denom)
          if spin == 1: spinor = -np.sqrt((l+m)*denom)
  else:
    if rank == 0:
      print('Spinor - j and l not compatible')
    quit()

  return spinor

def clebsch_gordan ( nawf, sh_l, sh_j, spol ):
  import numpy as np
  # Transformation matrices from the | l m s s_z > basis to the
  # | j mj l s > basis in the l-subspace
  #
  l = 0
  Ul0 = np.zeros((2*(2*l+1),2*(2*l+1)), dtype=float)
  Ul0[0,1] = 1
  Ul0[1,0] = 1

  l = 1
  Ul1 = np.zeros((2*(2*l+1),2*(2*l+1)), dtype=float)
  j = l - 0.5
  for m1 in range(1, 2*l+1):
      m = m1 - l
      Ul1[m1-1,2*(m1-1)+1-1] = spinor(l,j,m,0)
      Ul1[m1-1,2*(m1-1)+4-1] = spinor(l,j,m,1)
  j = l + 0.5
  for m1 in range(1, 2*l+2+1):
      m = m1 - l - 2
      if (m1 == 1):
         Ul1[m1+2*l-1,2*(m1-1)+2-1] = spinor(l,j,m,1)
      elif (m1==2*l+2):
         Ul1[m1+2*l-1,2*(m1-1)-1-1] = spinor(l,j,m,0)
      else:
         Ul1[m1+2*l-1,2*(m1-1)-1-1] = spinor(l,j,m,0)
         Ul1[m1+2*l-1,2*(m1-1)+2-1] = spinor(l,j,m,1)

  l = 2
  Ul2 = np.zeros((2*(2*l+1),2*(2*l+1)), dtype=float)
  j = l - 0.5
  for m1 in range(1, 2*l+1):
      m = m1 - l
      Ul2[m1-1,2*(m1-1)+1-1] = spinor(l,j,m,0)
      Ul2[m1-1,2*(m1-1)+4-1] = spinor(l,j,m,1)
  j = l + 0.5
  for m1 in range (1, 2*l+2+1):
      m = m1 - l - 2
      if (m1 == 1):
         Ul2[m1+2*l-1,2*(m1-1)+2-1] = spinor(l,j,m,1)
      elif (m1==2*l+2):
         Ul2[m1+2*l-1,2*(m1-1)-1-1] = spinor(l,j,m,0)
      else:
         Ul2[m1+2*l-1,2*(m1-1)-1-1] = spinor(l,j,m,0)
         Ul2[m1+2*l-1,2*(m1-1)+2-1] = spinor(l,j,m,1)

  l = 3
  Ul3 = np.zeros((2*(2*l+1),2*(2*l+1)), dtype=float)
  j = l - 0.5
  for m1 in range(1, 2*l+1):
      m = m1 - l
      Ul3[m1-1,2*(m1-1)+1-1] = spinor(l,j,m,0)
      Ul3[m1-1,2*(m1-1)+4-1] = spinor(l,j,m,1)
  j = l + 0.5
  for m1 in range (1, 2*l+2+1):
      m = m1 - l - 2      
      if (m1 == 1):
        Ul3[m1+2*l-1,2*(m1-1)+2-1] = spinor(l,j,m,1)
      elif (m1==2*l+2):
        Ul3[m1+2*l-1,2*(m1-1)-1-1] = spinor(l,j,m,0)
      else:
        Ul3[m1+2*l-1,2*(m1-1)-1-1] = spinor(l,j,m,0)
        Ul3[m1+2*l-1,2*(m1-1)+2-1] = spinor(l,j,m,1)


  Ul = [Ul0,Ul1,Ul2,Ul3]

  Ul1_alt=np.roll(np.roll(Ul1,-2,axis=0),-2,axis=1)
  Ul2_alt=np.roll(np.roll(Ul2,-4,axis=0),-4,axis=1)
  Ul3_alt=np.roll(np.roll(Ul3,-6,axis=0),-6,axis=1)

  Ul_alt = [Ul0,Ul1_alt,Ul2_alt,Ul3_alt]  

  # Build the full transformation matrix

  occ = [2,6,10,14]

  ntot = 0
  for n in range(len(sh_l)):
      ntot += occ[sh_l[n]]
  if ntot != nawf:
    raise ValueError('Wrong number of shells when creating \'Sj\'')

  Tn = np.zeros((ntot,ntot), dtype=float)

  
  n = 0
  for l in range(len(sh_l)):
    if sh_l[l]==0:
      Tn[n:n+occ[sh_l[l]],n:n+occ[sh_l[l]]] = Ul[sh_l[l]]
    else:
      if (sh_l[l]-sh_j[l])>0.0:
        # the case when j are sorted in ascending order in the pseudo
        Tn[n:n+occ[sh_l[l]],n:n+occ[sh_l[l]]] = Ul[sh_l[l]]
      else:
        # the case when j are sorted in descending order in the pseudo
        Tn[n:n+occ[sh_l[l]],n:n+occ[sh_l[l]]] = Ul_alt[sh_l[l]]

    n += occ[sh_l[l]]

  # Pauli matrices (x,y,z) 
  sP=0.5*np.array([[[0.0,1.0],[1.0,0.0]],[[0.0,-1.0j],[1.0j,0.0]],[[1.0,0.0],[0.0,-1.0]]])

  # Spin operator matrix  in the basis of |l,m,s,s_z>
  Sl = np.zeros((nawf,nawf), dtype=complex)
  for i in range(0, nawf, 2):
      Sl[i,i] = sP[spol][0,0]
      Sl[i,i+1] = sP[spol][0,1]
      Sl[i+1,i] = sP[spol][1,0]
      Sl[i+1,i+1] = sP[spol][1,1]

  # Spin operator matrix  in the basis of |j,m_j,l,s>
  Sj = np.zeros((nawf,nawf), dtype=complex)
  Sj = np.dot(Tn, Sl).dot(Tn.T)

  return Sj
